#include <avr/io.h>
#include <avr/interrupt.h> 
#include <util/delay.h> 
#include "dc_motor.h" 
#include "integer.h"  
#include "xprintf.h" 
#include "fifo.h"

fifo_t fifo; 

void KonfigUART()  // UART beallitasa
{
          UBRR0L = UBRR_ERTEK;        // UBRR_ERTEK also 8 bitjenek betoltese az UBRRL regiszterbe
          UBRR0H = (UBRR_ERTEK >> 8);   // UBRR_ERTEK felso 8 bitjenek betoltese az UBRRH regiszterbe
          // Aszinkron mod, 8 Adat Bit, Nincs Paritas Bit, 1 Stop Bit
          UCSR0C=(1 << UCSZ01)|(1 << UCSZ00); 
          UCSR0B |= (1 << RXEN0)|(1 << RXCIE0)|(1 << TXEN0);			//Ado aramkorok bekapcsolasa
		  UCSR0A|= (1 << U2X0);			// 2x sebessg belltsa		
}


void UARTAdatKuld(char data) // Ez a fuggveny a kuldendo adatot beirja az UDR regiszter kimeno pufferjebe
{
          while(!(UCSR0A & (1<<UDRE0)))  // Varakozas amig az Ado kesz nem lesz az adatkuldesre
          {
             //  Varakozas
          }
          // Az Ado mar kesz az adatkuldesre, a kuldendo adatot a kimeno pufferjebe irjuk
          UDR0=data;
}

void port_init(void)
{

	DDRB|=(1<<PB2);						//PWM out OC1B a SERVO motorhoz PB2
	DDRB|=(1<<PB3);						//PWM out OC2 a DC motorhoz		PB3
	PORTB=0;		
	DDRC|=(1<<PC2)|(1<<PC1)|(1<<PC0);	//R-2R DAC
	PORTC=0;
	DDRD=(1 <<PD1);						// a Tx kimenet a tbbi bemenet
	PORTD=0;
	
	M1A_DDR|=M1A;
	M1B_DDR|=M1B;
	MOTOR_STOP();
	//Analg kompartor belltsa
	ACSR|=(1<<ACIE)|(1<<ACI)|(1<<ACIS1)|(1<<ACIS0);
}


void interrupt_init(void)
{
	//DC motor PWM bellts			
	TCCR2A|=(1<<COM2A1)|(1<<WGM20);  	//fzishelyes PWM, 8-bit TOP=0xFF 
										//ha N=1 (CS20) akkor a PWM fr.=15.686Hz
										//ha N=8 (CS21), akkor a PWM fr.=8MHz/(N*510)=1.960Hz
	TCCR2B|=(1<<CS22)|(1<<CS21);		//ha N=32 (CS21|CS20) akkor a PWM fr.=490Hz
										//ha N=64 (CS22) akkor a PWM fr.=245Hz				
	TCNT2=0;							//ha N=128 (CS22|CS20) akkor a PWM fr.=245Hz				
  	OCR2A=0;							//--> ha N=256 (CS22|CS21) akkor a PWM fr.=61Hz  Ez a j
										//ha N=1024 (CS22|CS21|CS20) akkor a PWM fr.=15Hz
										
							
  	//SERVO-hoz bellts
	//Neknk 50Hz krli PWM frekvencia kell, ami 20msec peridusid
	// a SERVO-t 1-2 ms kztti impulzusokkal lehet vezrelni ez +-90fokos kitrst ad
	//ezzel ha 181 llst akarunk akkor 1ms/181-> akkor 5,52 usec felbonts kell
	//fast PWM-nl fpwm=fcpu/N*(1+TOP) itt N:1,8,64,256,1024 lehet.
	//Ha N=8 s a 14-es mdot hasznljuk ICR1 hatrozza meg a TOP rtkt, akkor ICR1=19999 addik
	//Ilyenkor a felbonts 1usec, teht az OCR1B rtknek 1-gyel trtn megvltoztatsa 1us-kel vltoztatja meg az imp hosszt
	//1ms hosszhoz OCR1B=1000; mg 2ms hosszhoz OCR1B=200 tartozik, kzplls 1500

	// ez az elmlet, a gyakorlatban 1600 a kzp 600 a min s 2600 a max lls
	//MEGA88 16. lb OCR1B,PB2
	
	TCCR1A=(1<<WGM11)|(1<<COM1A1)|(1<<COM1B1);
  	TCCR1B=(1<<WGM12)|(1<<WGM13)|(1<<CS11);	
	ICR1=19999;
  	OCR1B=SERVO_MID;	//SERVO

	TCCR0A=(1<<WGM01);										//MODE 2	CTC  
	TCCR0B=(1<<CS02)|(1<<CS00);								// oszts 1024
	TCNT0=0;
	OCR0A=77;												// oszt 77+1=78, ezzel 10ms-knt int	
	TIMSK0|=(1<<OCIE0A);

  	sei();		
}


ISR(USART_RX_vect)  // RX interrupt, ezek az utasitasok futak le ha adat erkezik a TX vonalon/
{
	rx_data = UDR0;
	_inline_fifo_put(&fifo, rx_data);
}

BYTE gp_comp (const char *str1, const prog_char *str2)
{
	char c;
	do 
	{
		c = pgm_read_byte(str2++);
	} while (c && (c == *str1++));
	return c;
}

BYTE check_sentences(const char* str)
{
	BYTE tipus = 0;
	if (gp_comp(str, PSTR("#STOP"))) 
		if (gp_comp(str, PSTR("#ON"))) 
			if (gp_comp(str, PSTR("#SS"))) 
				if (gp_comp(str, PSTR("#SL"))) 
					if (gp_comp(str, PSTR("#SR"))) 
						if (gp_comp(str, PSTR("#MS"))) 
							if (gp_comp(str, PSTR("#MF"))) 
								if (gp_comp(str, PSTR("#MR"))) 
									if (gp_comp(str, PSTR("#A_UC"))) 
										if (gp_comp(str, PSTR("#A_MO")))  
											if (gp_comp(str, PSTR("#DAC"))) 
												if (gp_comp(str, PSTR("#I_MO"))) 
													if (gp_comp(str, PSTR("#LA"))) return tipus; else tipus = LAMPA;
												else tipus = CURRENT_MOTOR;
											else tipus = DAC;
										else tipus = AKKU_MOTOR_WORD;
									else tipus = AKKU_UC_WORD;	
								else  tipus = DC_MOTOR_REW; 
							else tipus = DC_MOTOR_FOR;						
						else tipus = DC_MOTOR_STOP;						
					else tipus = SERVO_RIGHT;
				else tipus = SERVO_LEFT;
			else tipus = SERVO_STOP;	
		else tipus = AUTO_ON;	
	else tipus = AUTO_STOP;	
	return tipus;
}



int main()
{	
	int cc;
	char c;
	BYTE i=0;
	char buffer_[BLUE_SENTENCES_BUFFER];	
	
	//OSCCAL=148;
	port_init();
	KonfigUART();
	fifo_init(&fifo,f_buffer,MAX_FIFO_BUFFER);
	interrupt_init();
	//FEK_LAMP_ON;
	while(1)
	{ 
		if(!BLUE_CONNECT()) {
			OCR1B=SERVO_MID;
			OCR2A=0;
			RL_REFLEKTOR_OFF;
		} 
		while((cc = fifo_get_nowait(&fifo)) != -1)
		{	
			c=cc;	
			if(c == ',')
			{
				buffer_[i]='\0';
				switch(check_sentences(buffer_))
				{
					case SERVO_STOP : 
					{	
						OCR1B=SERVO_MID;
						break;
					}
					case SERVO_LEFT : 
					{	
						BYTE j=3;
						int x=0;
						while(buffer_[j])
						{
							x=10*x+(buffer_[j]-48);
							j++;
						}
						if(x <= 100)
						{
							OCR1B=SERVO_MID+10*x;
						} 
						break;
					}
					case SERVO_RIGHT : 
					{	
						BYTE j=3;
						int x=0;
						while(buffer_[j])
						{
							x=10*x+(buffer_[j]-48);
							j++;
						}	
						if(x <= 100)
						{
							OCR1B=SERVO_MID-10*x;
						} 
						break;
					}
					case DC_MOTOR_STOP : 
					{	
						MOTOR_STOP();
						OCR2A=0;
						FEK_LAMP_OFF;
						break;
					}
					case DC_MOTOR_REW : 
					{	
						FEK_LAMP_OFF;
						BYTE j=3,x=0;
						while(buffer_[j])
						{
							x=10*x+(buffer_[j]-48);
							j++;
						}
						MOTOR_HATRA();
						x=3*(x/2);
						if(x < OCR2A) {FEK_LAMP_ON;fek_timer=70;} else {FEK_LAMP_OFF;}
						OCR2A=x;
						pOCR=x;
						break;
					}
					case DC_MOTOR_FOR : 
					{	
						FEK_LAMP_OFF;
						BYTE j=3,x=0;
						while(buffer_[j])
						{
							x=10*x+(buffer_[j]-48);
							j++;
						}
						MOTOR_ELORE();
						x=10*(x/4);
						if(x < OCR2A) {FEK_LAMP_ON;fek_timer=70;} else {FEK_LAMP_OFF;}
						OCR2A=x;
						pOCR=x;
						break;
					}
					case AKKU_UC_WORD :
					{
						xsprintf(buf,"#AKKU_UC;%03d\n\0",akku_uc);
						BYTE i=0;
						while((buf[i]) && (i<31))
						{
							UARTAdatKuld(buf[i++]);
						}	
						break;
					}
					case AKKU_MOTOR_WORD :
					{
						xsprintf(buf,"#AKKU_MOTOR;%03d\n\0",akku_motor);
						BYTE i=0;
						while((buf[i]) && (i<31))
						{
							UARTAdatKuld(buf[i++]);
						}	
						break;
					}
					case CURRENT_MOTOR :
					{
						xsprintf(buf,"#CURR_MOT;%03d\n\0",motor_current/curr_db);
						BYTE i=0;
						while((buf[i]) && (i<31))
						{
							UARTAdatKuld(buf[i++]);
						}	
						curr_db=0;
						motor_current=0;
						cm_timer=150;
						break;
					}
					case DAC : 
					{	
						DAC_OUT((buffer_[4]-48));
						
						break;
					}
					case LAMPA : 
					{	
						lampa=(buffer_[3]-48);
						if(lampa&1) {LEFT_REFLEKTOR_ON;} else {LEFT_REFLEKTOR_OFF;}
						if(lampa&2) {RIGHT_REFLEKTOR_ON;} else {RIGHT_REFLEKTOR_OFF;}
						break;
					}
					case AUTO_STOP : 
					{	
						OCR2A=0;
						pOCR=0;
						OCR1B=SERVO_MID;
						stop=TRUE;
						break;
					}	
					case AUTO_ON : 
					{	
						if(lampa&1) {LEFT_REFLEKTOR_ON;} else {LEFT_REFLEKTOR_OFF;}
						if(lampa&2) {RIGHT_REFLEKTOR_ON;} else {RIGHT_REFLEKTOR_OFF;}
						stop=FALSE;
						break;
					}
				}
				buffer_[i++]='\r';
				buffer_[i++]='\n';
				buffer_[i++]='\0';
				i=0;
			} 
			else 
			{
				buffer_[i]=c;
				if(i < BLUE_SENTENCES_BUFFER-1) i++;
			}	
		}
	}	
}

ISR(TIMER0_COMPA_vect)
{
	static BYTE szam;
	static BYTE _stop;
	
	if(stop)
	{
		_stop++;
		if(!(_stop%100))
		{
			RL_REFLEKTOR_CHANGE;
			_stop=0;
		}
	}
	szam++;	
	if(!(szam%10)) 		//100ms-enknt megnzi az akku feszltsgeket
	{
		ADMUX =(1<<REFS1)|(1<<REFS0)|(1<<ADLAR)|AKKU_MOTOR;				// 1,1V a referencia fesz,  balra igaztott 8 bit
		ADCSRA= (1<<ADEN)|(1<<ADSC)|(1<<ADIE)|(1<<ADPS2)|(1<<ADPS1); 	//ADC eng, oszt 64, Konverzi indts, ADC interrupt eng
		szam=0;
	}
	if(!BLUE_CONNECT()) {FEK_LAMP_ON;}
	if(analog_timer) {analog_timer--;}	else		//vrunk az ram cskkensre
	{
		
		if(pOCR > OCR2A) delta=pOCR-OCR2A; else delta=0;
		if(ANALOG_COMP_ENABLED)		//ha engedlyezve van a komp ISR akkor mr lehet nvelni OCR2A-t
		{						//mert, ha nem akkor szlna a komp ISR
			if(delta) OCR2A+=delta/2;
		} else {ACSR|=(1<<ACIE);OCL_LED_OFF;}
	}
	if(cm_timer) {cm_timer--;}	else {
		OCR2A=0;
		pOCR=0;
		OCR1B=SERVO_MID;
		stop=TRUE;
	}
	if(!stop){
		if(fek_timer) {fek_timer--;}	else {
			FEK_LAMP_OFF;
		}
		
	}
}


ISR(ADC_vect)
{
	BYTE adat,n=0b00000111;
	adat = ADCH;
	switch (ADMUX & 0b00001111) 
	{
		case AKKU_MOTOR :
		{
			akku_motor=adat;
			ADMUX = (1<<REFS1)|(1<<REFS0)|(1<<ADLAR)|BLUE_CONN;	// s betesszk az AKKU lekrdezsnek megfelelt.
			n = 0b11001110;								//(1<<ADEN) | (1<<ADSC) | (1<<ADIE) | (1<<ADPS2)| (1<<ADPS1)
			break;
		}
		case BLUE_CONN :
		{
			blue_conn=adat;	
			ADMUX = (1<<REFS1)|(1<<REFS0)|(1<<ADLAR)|MOTOR_CURRENT;		// s betesszk a MOTOR_CURRENT lekrdezsnek megfelelt.
			n = 0b11001110;
			break;
		}	
		case MOTOR_CURRENT :
		{
			if(adat < 5) adat=0;
			motor_current+=adat;	
			curr_db++;
			ADMUX = (1<<REFS1)|(1<<REFS0)|(1<<ADLAR)|AKKU_UC;		// s betesszk a AKKU_UC lekrdezsnek megfelelt.
			n = 0b11001110;
			break;
		}	
		case AKKU_UC :
		{
			akku_uc=adat;	
			n = 0;											//konverzi vge, legkzelebb, majd a Timer0COMPA ISR-ben lesz jraindtva
			break;
		}	
	}
	ADCSRA 	= n;
}

ISR(ANALOG_COMP_vect)
{
	ACSR&=~(1<<ACIE);
	analog_timer=90;
	OCL_LED_ON;
	if(OCR2A > DELTA) OCR2A-=DELTA;
}
